from uuid import uuid4
import getpass
import logging
import sys


class Service:
    '''
    Represents a service.

    This can also be used to represent a system program,
    since a program is also a service.

    All services which inherits from this class
    will be automatically initialized.

    Put the order of the services to initialize in
    the `service_init_order` attribute.

    Put the short name of the service which should not be
    initialized in the `disable_services` attribute.
    '''

    # This can be overridden by child classes to change the service name.
    NAME = 'Core'

    # For the yes-no user prompt
    OPTION_YES = 'y'
    OPTION_NO = 'n'

    DEPENDENCIES = 'dependencies'
    SERVICE_TYPE = 'service_type'
    SERVICE_SUFFIX = 'service'

    # Enable to show service initialization progress.
    SHOW_SERVICE_INIT_LOG = False

    # Order of initializing services.
    service_init_order = [
    ]

    # Definition of services goes here.
    services = {
    }

    # List of services to remain uninitialized.
    disable_services = [
    ]

    @classmethod
    def msg(cls, *args, **kwargs):
        '''
        Prints a message to stdout.
        '''
        return print(*args, **kwargs)

    @classmethod
    def main(cls):
        '''
        This should be overridden by child classes.
        '''
        pass

    @classmethod
    def generate_id(cls):
        return str(uuid4())

    def new_id(self):
        id = self.generate_id()
        while self.get(id):
            id = self.generate_id()
        return id

    def new(self):
        '''
        Returns a new object of the type the service is representing.

        Child classes can override this.
        '''
        return None

    def get(self, id):
        '''
        Get the object the service is representing.

        Child classes can override this.
        '''
        return None

    def remove(self, obj):
        '''
        Remove the object the service is representing.

        Child classes can override this.
        '''
        pass

    def add(self, obj):
        '''
        Adds or update the object the service is representing.

        Child classes can override this.
        '''
        pass

    def listing(self, page=None):
        '''
        Returns a list of the objects the service is representing.

        Child classes can override this.
        '''
        return []

    def exit(self):
        sys.exit()

    def user_prompt(self, prompt, password=False, yes_no=False,
            yes_no_default=None):
        '''
        If Yes-No mode is enabled, returns either True or False.
        True corresponds to 'Yes' and False corresponds to 'False'.
        '''
        prompt += ' '
        if password:
            ret = getpass.getpass(prompt)
        else:
            ret = input(prompt)
        if yes_no:
            if not yes_no_default:
                yes_no_default = self.OPTION_NO
            ret = ret.strip().lower()
            if ret == '':
                ret = yes_no_default
            if ret == self.OPTION_YES:
                return True
            else:
                return False
        return ret

    def get_service_attribute_name(self, srv_name):
        '''
        Returns the proper service's attribute name.

        Example:
        self.service_attribute_name('core') => 'core_service'
        '''
        n = '%s_%s' % (
            srv_name,
            self.SERVICE_SUFFIX
            )
        return n

    def set_service(self, srv_name, obj):
        n = self.get_service_attribute_name(srv_name)
        setattr(self, n, obj)

    def get_service(self, srv_name):
        n = self.get_service_attribute_name(srv_name)
        return getattr(self, n)

    def set_service_dependencies(self, srv_name):
        if srv_name not in self.services:
            return
        srv_def = self.services[srv_name]
        if self.DEPENDENCIES not in srv_def:
            return
        dependencies = srv_def[self.DEPENDENCIES]
        srv_obj = self.get_service(srv_name)
        if not srv_obj:
            return
        for dep_srv_name in dependencies:
            name = self.get_service_attribute_name(dep_srv_name)
            dep_srv = self.get_service(dep_srv_name)
            setattr(srv_obj, name, dep_srv)

    def initialize_service(self, srv_name):
        srv_obj = self.get_service(srv_name)
        try:
            srv_obj.initialize()
            if self.SHOW_SERVICE_INIT_LOG:
                self.msg('Service initialized: %s' % (srv_obj.NAME,))
        except:
            self.msg('Service initialization error: %s' % (srv_obj.NAME,))
            raise

    def initialize(self):
        '''
        Automatically initializes defined services.

        Define services to initialize in the services list.

        Example:
        If you define 'core' in the services list, the 'init_core_service'
        method will be executed. Its return value is then assigned to the
        attribute 'core_service'.
        '''
        for k, v in self.services.items():
            type_object= v[self.SERVICE_TYPE]
            if not callable(type_object):
                raise ValueError('Service class must be specified.')
            srv_obj = type_object()
            if k not in self.disable_services:
                self.set_service(k, srv_obj)
        for srv_name in self.services:
            self.set_service_dependencies(srv_name)
        for srv_name in self.service_init_order:
            self.initialize_service(srv_name)