from datetime import datetime
from . import json
from .redis import Redis


class StoreObject:
    def __init__(self, obj=None):
        self.created_at = datetime.now()
        self.updated_at = datetime.now()
        if obj:
            self.__dict__ = obj

    def __getattr__(self, attr_name):
        return ''

    @property
    def primary_attr(self):
        return None

    def convert(self):
        'Convert attributes to the proper data types.'
        pass

    def get_list_as_field_value(self, l, separator='\r\n'):
        return separator.join(l)

    def get_field_value_as_list(self, req, field_name, separator='\r\n'):
        s = req.get_body_argument(field_name)
        s = s.strip()
        if not s:
            return []
        toks = s.split(separator)
        return toks

    def validate(self):
        "Returns a list of validation error messages. Empty list if all ok."
        return []


class Store:
    '''
    Representing a store that works with Redis to store its content.
    '''
    PREFIX = 'store'
    ATTRIBUTE_NAME = 'name'
    STORE_OBJ_TYPE = StoreObject

    def __init__(self):
        self.db = None
        self.config = None

    def connect(self, config_obj):
        self.config = config_obj
        host = config_obj.redis['host']
        port = config_obj.redis['port']
        self.db = Redis.connect(host, port)

    def get(self, name):
        '''
        Get content by name.
        '''
        name = self.get_name(name)
        obj = self.db.get(name)
        if obj:
            obj = obj.decode()
            obj = json.loads(obj)
            s = self.STORE_OBJ_TYPE(obj)
            s.convert()
            return s
        return None

    def get_name(self, name):
        return '{0}:{1}'.format(self.PREFIX, name)

    def listing(self):
        prefix_colon = '{0}:'.format(self.PREFIX)
        l = self.db.listing('{0}*'.format(prefix_colon))
        def remove_prefix(entry):
            entry = entry.decode()
            has_prefix = entry.startswith(prefix_colon)
            if has_prefix:
                return entry[len(prefix_colon):]
            return entry
        def convert_to_object(entry):
            return self.get(entry)
        l = map(remove_prefix, l)
        l = map(convert_to_object, l)
        return list(l)

    def new(self):
        return self.STORE_OBJ_TYPE()

    def remove(self, obj):
        self.remove_by_primary_attr(obj.primary_attr)

    def remove_by_primary_attr(self, pri_attr):
        name = self.get_name(pri_attr)
        self.db.remove(name)

    def save(self, obj, orig_pri_attr=None, update_timestamp=True, expires=None):
        if orig_pri_attr and obj.primary_attr != orig_pri_attr:
            self.remove_by_primary_attr(orig_pri_attr)
        if getattr(obj, 'updated_at') and update_timestamp:
            obj.updated_at = datetime.now()
        content = json.dumps(obj)
        name = getattr(obj, self.ATTRIBUTE_NAME)
        name = self.get_name(name)
        return self.db.set(name, content, expires)