from datetime import datetime, timedelta
from decimal import Decimal
from tornado.ioloop import IOLoop
from tornado.web import Application as BaseApplication
from tornado.web import MissingArgumentError
from tornado.web import RequestHandler as BaseRequestHandler
from tornado.web import UIModule
from tornado.web import URLSpec
from tornado.web import StaticFileHandler
from tornado.httpserver import HTTPServer
from uuid import uuid4
from .store import Store, StoreObject
from . import json
import tornado.web


YES = 'yes'


def namespace(name):
    '''
    Decorator. Adds namespace to decorated class.
    '''
    def namespaced_class(cls):
        if cls.__name__ != 'NamespacedClass':
            original_name = '%s.%s' % (
                cls.__module__,
                cls.__name__,
                )
            cls._original_name_ = original_name
        class NamespacedClass(cls, metaclass=Meta):
            @classmethod
            def route(cls):
                r_pattern = '/%s%s'
                r = super().route()
                if isinstance(r, list):
                    out = []
                    for route in r:
                        route = r_pattern % (name, route)
                        out.append(route)
                else:
                    out = r_pattern % (name, r)
                return out

            def template_filename(self):
                t = super().template_filename()
                t = '%s/%s' % (name, t)
                return t
        return NamespacedClass
    return namespaced_class


class Meta(type):
    def __repr__(cls):
        route = cls.route()
        out = '<class %s namespaced %s>' % (
            repr(cls._original_name_),
            repr(route),
            )
        return out


class Application(BaseApplication):
    '''
    Represents the web application.

    This utilises a tornado.web.Application object.
    '''

    def __init__(self, request_handlers, additional_routes=[], default_host='',
            transforms=None, **settings):
        '''
        Arguments:
        handlers -- list of RequestHandler objects
        additional_routes -- additional routes. Specify like how it is done in
                             Tornado Web Framework
        default_host -- similar to 'default_host' parameter in
                        tornado.web.Application
        transforms -- similar to 'transforms' parameter in
                      tornado.web.Application
        settings -- similar to 'settings' parameter in tornado.web.Application
        '''
        self.http_server = None
        self.ioloop_instance = None
        self.request_handlers = request_handlers
        self.additional_routes = additional_routes
        self.http_servers = []
        routes = self.generate_route_list()
        super().__init__(routes, default_host, transforms, **settings)

    def generate_route_list(self):
        out_lst = []
        for handler in self.request_handlers:
            route = handler.route()
            if isinstance(route, list):
                for r in route:
                    t = URLSpec(r, handler, name=r)
                    out_lst.append(t)
            elif isinstance(route, str):
                t = URLSpec(route, handler, name=route)
                out_lst.append(t)
        out_lst += self.additional_routes
        return out_lst

    def listen(self, port, address='', **kwargs):
        http_server = HTTPServer(self, **kwargs)
        http_server.bind(port, address)
        self.http_servers.append(http_server)

    def start(self):
        '''
        Convenience method for starting the IO loop.
        '''
        for http_server in self.http_servers:
            http_server.start()
        self.ioloop_instance = IOLoop.current()
        return self.ioloop_instance.start()

    def stop(self):
        '''
        Convenience method for stopping the IO loop.
        '''
        self.ioloop_instance.stop()


class RequestHandler(BaseRequestHandler):
    MIME_TYPE_JSON = 'application/json'
    MIME_TYPE_TEXT = 'text/plain'

    @classmethod
    def route(cls):
        '''
        Route can be a string or a list of strings.
        The route string is identical to how routes are specified in Tornado
        Web Framework.

        Child classes should override this.
        '''
        return NotImplementedError

    def redirect_url(self):
        '''
        Returns the URL to redirect to.

        Child classes should override this.
        '''
        return None

    def template_filename(self):
        '''
        Returns the template filename. This is used as a pathname to load the template files.

        Child classes can override this.
        '''
        return None

    def __init__(self, application, request, **kwargs):
        super().__init__(application, request, **kwargs)
        self.template_context = {}

    def error_not_found(self):
        self.set_status_with_message(404, 'Not found')

    def error_500(self):
        self.set_status_with_message(500, 'Internal Server Error')

    def get_body_argument_as_decimal(self, name, default=[], strip=True):
        value = self.get_body_argument(name, default, strip)
        if not value.strip():
            return None
        return Decimal(value)

    def is_body_argument_yes(self, name, default=None, strip=True):
        value = self.get_body_argument(name, default, strip)
        if isinstance(value, str) and value.strip().lower() == YES:
            return True
        return False

    def is_http_get(self):
        return self.is_http_method('GET')

    def is_http_method(self, method):
        return self.request.method == method.upper()

    def is_http_post(self):
        return self.is_http_method('POST')

    def template_context_key(self, k):
        if k not in self.template_context:
            return None
        return self.template_context[k]

    def redirect(self, url=''):
        if not url:
            url = self.redirect_url()
        super().redirect(url)

    def redirect_to_handler(self, handler_type, *args):
        '''
        Redirects to the specified request handler.
        '''
        url = self.reverse_url_for_handler(handler_type, *args)
        return self.redirect(url)

    def reverse_url_for_handler(self, handler_type, *args):
        r = handler_type.route()
        if isinstance(r, list):
            r = r[0]
        url = self.reverse_url(r, *args)
        return url

    def template_render(self):
        '''
        Similar to tornado.web.RequestHandler#render.
        '''
        f_name = self.template_filename()
        if f_name is None:
            return
        return self.render(f_name, **self.template_context)

    def template_render_string(self, template_name, **kwargs):
        '''
        Similar to tornado.web.RequestHandler#render_string
        '''
        f_name = self.template_filename()
        if f_name is None:
            return
        self.set_flash_messages_to_template_context()
        return self.render(f_name, **self.template_context)

    def set_content_type(self, mime_type, append_charset_utf8=True):
        c = mime_type
        if append_charset_utf8:
            c += '; charset=utf-8'
        return self.set_header('Content-Type', c)

    def set_status_with_message(self, code, message):
        self.set_status(code, message)
        self.set_content_type(self.MIME_TYPE_TEXT)
        self.write(message)

    def set_template_context_key(self, k, v):
        self.template_context[k] = v

    def write_json(self, obj, json_lib=None):
        if json_lib:
            out = json_lib.dumps(obj)
        else:
            out = json.dumps(obj)
        self.set_content_type(self.MIME_TYPE_JSON)
        self.write(out)


class LoginSession(StoreObject):
    def __init__(self, obj=None):
        self.id = ''
        self.msgs = []
        self.since = datetime.now()
        super().__init__(obj)


class LoginSessionStore(Store):
    PREFIX = 'login-session'
    ATTRIBUTE_NAME = 'id'
    STORE_OBJ_TYPE = LoginSession

    def clear_expired_sessions(self):
        l = self.listing()
        now = datetime.now()
        for s_obj in l:
            if now - s_obj.since >= timedelta(days=1):
                self.remove(s_obj)

    def generate_id(self):
        return str(uuid4())

    def new_id(self):
        id = self.generate_id()
        while self.get(id):
            id = self.generate_id()
        return id

    def new(self):
        obj = self.STORE_OBJ_TYPE()
        obj.id = self.new_id()
        return obj

    def save(self, obj, orig_pri_attr=None):
        obj.since = datetime.now()
        super().save(obj, orig_pri_attr)